from __future__ import annotations

import sys
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Iterator, NamedTuple, cast

from xsdata.formats.dataclass.filters import Filters
from xsdata.formats.dataclass.generator import DataclassGenerator

from ome_autogen import _util
from xsdata_pydantic_basemodel.generator import PydanticBaseFilters

if TYPE_CHECKING:
    from jinja2 import Environment, FileSystemLoader
    from xsdata.codegen.models import Attr, Class
    from xsdata.codegen.resolver import DependenciesResolver
    from xsdata.models.config import GeneratorConfig


# from ome_types._mixins._base_type import AUTO_SEQUENCE
# avoiding import to avoid build-time dependency on the ome-types package
AUTO_SEQUENCE = "__auto_sequence__"

# Methods and/or validators to add to generated classes
# (predicate, method) where predicate returns True if the method should be added
# to the given class.  Note: imports for these methods are added in
# IMPORT_PATTERNS below.
ADDED_METHODS: list[tuple[Callable[[Class], bool], str]] = [
    (
        lambda c: c.name == "BinData",
        "\n\n_v = root_validator(pre=True)(bin_data_root_validator)",
    ),
    (
        lambda c: c.name == "Value",
        "\n\n_v = validator('any_elements', each_item=True)(any_elements_validator)",
    ),
    (
        lambda c: c.name == "Pixels",
        "\n\n_v = root_validator(pre=True)(pixels_root_validator)",
    ),
    (
        lambda c: c.name == "XMLAnnotation",
        "\n\n_v = validator('value', pre=True)(xml_value_validator)",
    ),
    (
        lambda c: c.name == "PixelType",
        "\n\nnumpy_dtype = property(pixel_type_to_numpy_dtype)",
    ),
]


class Override(NamedTuple):
    element_name: str  # name of the attribute in the XSD
    class_name: str  # name of our override class
    module_name: str | None  # module where the override class is defined


CLASS_OVERRIDES = [
    Override("FillColor", "Color", "ome_types.model._color"),
    Override("StrokeColor", "Color", "ome_types.model._color"),
    Override("Color", "Color", "ome_types.model._color"),
    Override("Union", "ShapeUnion", "ome_types.model._shape_union"),
    Override(
        "StructuredAnnotations",
        "StructuredAnnotationList",
        "ome_types.model._structured_annotations",
    ),
]
# classes that should never be optional, but always have default_factories
NO_OPTIONAL = {"Union", "StructuredAnnotations"}

# if these names are found as default=..., turn them into default_factory=...
FACTORIZE = set([x.class_name for x in CLASS_OVERRIDES] + ["StructuredAnnotations"])

# prebuilt maps for usage in code below
OVERRIDE_ELEM_TO_CLASS = {o.element_name: o.class_name for o in CLASS_OVERRIDES}
IMPORT_PATTERNS = {
    o.module_name: {
        o.class_name: [f": {o.class_name} =", f": Optional[{o.class_name}] ="]
    }
    for o in CLASS_OVERRIDES
    if o.module_name
}
IMPORT_PATTERNS.update(
    {
        "ome_types._mixins._util": {"new_uuid": ["default_factory=new_uuid"]},
        "datetime": {"datetime": ["datetime"]},
        "ome_types._mixins._validators": {
            "any_elements_validator": ["any_elements_validator"],
            "bin_data_root_validator": ["bin_data_root_validator"],
            "pixels_root_validator": ["pixels_root_validator"],
            "xml_value_validator": ["xml_value_validator"],
            "pixel_type_to_numpy_dtype": ["pixel_type_to_numpy_dtype"],
        },
    }
)


class OmeGenerator(DataclassGenerator):
    @classmethod
    def init_filters(cls, config: GeneratorConfig) -> Filters:
        return OmeFilters(config)

    def render_module(
        self, resolver: DependenciesResolver, classes: list[Class]
    ) -> str:
        mod = super().render_module(resolver, classes)

        # xsdata renames classes like "FillRule" (which appears as a SimpleType
        # inside of the Shape ComlexType) as "Shape_FillRule".
        # We want to make them available as "FillRule" in the corresponding
        # module, (i.e. the "Shape" module in this case).
        # That is, we want "Shape = Shape_FillRule" included in the module.
        # this is for backwards compatibility.
        aliases = []
        for c in classes:
            for i in resolver.imports:
                if f"{c.name}_" in i.qname:
                    # import_name is something like 'Shape_FillRule'
                    import_name = i.qname.rsplit("}", 1)[-1]
                    # desired alias is just 'FillRule'
                    alias = import_name.split(f"{c.name}_")[-1]
                    aliases.append(f"{alias} = {import_name}")

            # we also want inner (nested) classes to be available at the top level
            # e.g. map.Map.M -> map.M
            for inner in c.inner:
                aliases.append(f"{inner.name} = {c.name}.{inner.name}")

        if aliases:
            mod += "\n\n" + "\n".join(aliases) + "\n"

        return mod


class OmeFilters(PydanticBaseFilters):
    def register(self, env: Environment) -> None:
        # add our own templates dir to the search path
        tpl_dir = Path(__file__).parent.joinpath("templates")
        cast("FileSystemLoader", env.loader).searchpath.insert(0, str(tpl_dir))
        super().register(env)
        env.filters.update({"methods": self.methods})

    def __init__(self, config: GeneratorConfig):
        super().__init__(config)

        # TODO: it would be nice to know how to get the schema we're processing from
        # the config.  For now, we just assume it's the OME schema and that's the
        # hardcoded default in _util.get_appinfo
        self.appinfo = _util.get_appinfo()

    @contextmanager
    def _modern_typing(self) -> Iterator[None]:
        """Context manager to use modern typing syntax."""
        prev_u, self.union_type = self.union_type, True
        prev_s, self.subscriptable_types = self.subscriptable_types, True
        try:
            yield
        finally:
            self.union_type = prev_u
            self.subscriptable_types = prev_s

    def class_params(self, obj: Class) -> Iterator[tuple[str, str, str]]:
        # This method override goes along with the docstring jinja template override
        # to fixup the numpy docstring format.
        # https://github.com/tefra/xsdata/issues/818

        if sys.version_info < (3, 8):  # pragma: no cover
            # i don't know why, but in python 3.7, it's not picking up our
            # template ... so this method yields too many values to unpack
            # xsdata/formats/dataclass/templates/docstrings.numpy.jinja2", line 6
            #   {%- for var_name, var_doc in obj | class_params %}
            #       ValueError: too many values to unpack (expected 2)
            # however, it's not at all important that we be able to build docs correctly
            # on python 3.7.  The built wheel will still work fine (and won't be built
            # on python 3.7 anyway)
            yield from super().class_params(obj)
            return

        for attr in obj.attrs:
            name = attr.name
            name = (
                self.constant_name(name, obj.name)
                if obj.is_enumeration
                else self.field_name(name, obj.name)
            )
            with self._modern_typing():
                type_ = self.field_type(attr, [obj.name])
            help_ = attr.help
            if not help_:
                help_ = f"(The {obj.name} {attr.name})."
            yield name, type_, self.clean_docstring(help_)

    def class_bases(self, obj: Class, class_name: str) -> list[str]:
        # we don't need PydanticBaseFilters to add the Base class
        # because we add it in the config.extensions
        # This could go once PydanticBaseFilters is better about deduping
        return Filters.class_bases(self, obj, class_name)

    def _attr_is_optional(self, attr: Attr) -> bool:
        if attr.name in NO_OPTIONAL:
            return False
        return attr.is_nillable or (
            attr.default is None and (attr.is_optional or not self.format.kw_only)
        )

    def _format_type(self, attr: Attr, result: str) -> str:
        if self._attr_is_optional(attr):
            return f"None | {result}" if self.union_type else f"Optional[{result}]"
        return result

    def field_type(self, attr: Attr, parents: list[str]) -> str:
        if attr.is_list and not getattr(attr, "_plural_set", False):
            # HACK
            # It would be nicer to put this in the self.field_name method...but that
            # method only receives the attr name, not the attr object, and so we
            # don't know at that point whether it belongs to a list or not.
            # This hack works only because this method is called BEFORE self.field_name
            # in the class.jinja2 template, so we directly modify the attr object here.
            attr.name = self.appinfo.plurals.get(attr.name, f"{attr.name}s")
            attr._plural_set = True  # type: ignore

        if attr.name in OVERRIDE_ELEM_TO_CLASS:
            return self._format_type(attr, OVERRIDE_ELEM_TO_CLASS[attr.name])

        type_name = super().field_type(attr, parents)
        # we want to use datetime.datetime instead of XmlDateTime
        return type_name.replace("XmlDateTime", "datetime")

    @classmethod
    def build_import_patterns(cls) -> dict[str, dict]:
        patterns = super().build_import_patterns()
        patterns.setdefault("pydantic", {}).update(
            {
                "validator": ["validator("],
                "root_validator": ["root_validator("],
            }
        )
        patterns.update(IMPORT_PATTERNS)
        return {key: patterns[key] for key in sorted(patterns)}

    def field_default_value(self, attr: Attr, ns_map: dict | None = None) -> str:
        if attr.tag == "Attribute" and attr.name == "ID":
            return repr(AUTO_SEQUENCE)
        for override in CLASS_OVERRIDES:
            if attr.name == override.element_name:
                if not self._attr_is_optional(attr):
                    return override.class_name
        return super().field_default_value(attr, ns_map)

    def format_arguments(self, kwargs: dict, indent: int = 0) -> str:
        # keep default_factory at the front
        if kwargs.get("default") in FACTORIZE:
            kwargs = {"default_factory": kwargs.pop("default"), **kwargs}

        # uncomment this to use new_uuid as the default_factory for all UUIDs
        # but then we have an equality checking problem in the tests
        # if kwargs.get("metadata", {}).get("pattern", "").startswith("(urn:uuid:"):
        #     kwargs.pop("default", None)
        #     kwargs = {"default_factory": "new_uuid", **kwargs}
        return super().format_arguments(kwargs, indent)

    def constant_name(self, name: str, class_name: str) -> str:
        if class_name in self.appinfo.enums:
            # use the enum names found in appinfo/xsdfu/enum
            return self.appinfo.enums[class_name][name].enum
        return super().constant_name(name, class_name)

    def methods(self, obj: Class) -> list[str]:
        for predicate, code in ADDED_METHODS:
            if predicate(obj):
                return [code]
        return []



from __future__ import annotations

import io
import operator
import os
import warnings
from contextlib import nullcontext, suppress
from functools import lru_cache
from pathlib import Path
from struct import Struct
from typing import TYPE_CHECKING, cast, overload

from pydantic import BaseModel
from xsdata.formats.dataclass.parsers.config import ParserConfig

from xsdata_pydantic_basemodel.bindings import (
    SerializerConfig,
    XmlParser,
    XmlSerializer,
)

try:
    from lxml import etree as ET
except ImportError:  # pragma: no cover
    from xml.etree import ElementTree as ET  # type: ignore[no-redef]


if TYPE_CHECKING:
    from typing import Any, BinaryIO, ContextManager, TypedDict

    import xmlschema
    from lxml.etree import _XSLTResultTree
    from typing_extensions import Literal
    from xsdata.formats.dataclass.parsers.mixins import XmlHandler

    from ome_types._mixins._base_type import OMEType
    from ome_types.model import OME
    from xsdata_pydantic_basemodel.bindings import XmlContext

    XMLSource = Path | str | bytes | io.BytesIO
    FileLike = str | io.BufferedIOBase
    ElementOrTree = ET._Element | ET._ElementTree

    class ParserKwargs(TypedDict, total=False):
        config: ParserConfig
        context: XmlContext
        handler: type[XmlHandler]


__all__ = ["from_xml", "to_xml", "to_dict", "from_tiff", "tiff2xml"]

OME_ROOT = "http://www.openmicroscopy.org/Schemas/OME"
OME_2016_06_URI = f"{OME_ROOT}/2016-06"
OME_2016_06_NS = f"{{{OME_2016_06_URI}}}"
OME_2016_06_XSD = str(Path(__file__).parent / "ome-2016-06.xsd")


def from_xml(
    source: XMLSource,
    *,
    validate: bool | None = None,
    parser: Any = None,
    parser_kwargs: ParserKwargs | None = None,
    warn_on_transform: bool = False,
) -> OME:  #  Not totally true, see note below
    """Generate an OME object from an XML document.

    NOTE: Technically, this can return any ome-types instance, (not just OME) but it's
    by far the most common type that will come out of this function, and the type
    annotation will be more useful to most users. For those who pass in an xml document
    that isn't just a root <OME> tag, they can cast the result to the correct type
    themselves.

    Parameters
    ----------
    source : Path | str | bytes | io.BytesIO,
        Path to an XML file, string or bytes containing XML, or a file-like object.
        If the source is not OME-2016-06 XML, it will be transformed to that namespace
        if possible.
    validate : bool | None
        Whether to validate the XML document against the OME schema.
        If None, validation will be skipped if lxml is not available,
        and will be performed otherwise.
    parser : Any
        Ignored, but kept for backwards compatibility.
    parser_kwargs : ParserKwargs | None
        Passed to the XmlParser constructor. If None, a default parser
        will be used.
    warn_on_transform : bool
        Whether to warn if a transformation was applied to bring the document to
        OME-2016-06.

    Returns
    -------
    OME
        The OME object parsed from the XML document. (See NOTE above.)
    """
    if parser is not None:  # pragma: no cover
        warnings.warn(
            "As of version 0.4.0, the parser argument is ignored. "
            "lxml will be used if available in the environment, but you can "
            "drop this keyword argument.",
            DeprecationWarning,
            stacklevel=2,
        )

    if validate:
        xml_2016 = validate_xml(source, warn_on_transform=warn_on_transform)
    else:
        xml_2016 = ensure_2016(
            source, warn_on_transform=warn_on_transform, as_tree=True
        )

    OME_type = _get_root_ome_type(xml_2016)
    parser = XmlParser(**(parser_kwargs or {}))
    return parser.parse(xml_2016, OME_type)


# ------------------------


def from_tiff(
    path: Path | str | BinaryIO,
    *,
    validate: bool | None = None,
    parser_kwargs: ParserKwargs | None = None,
) -> OME:
    """Generate an OME object from a TIFF file.

    Parameters
    ----------
    path : Path | str | BinaryIO
        Path to a TIFF file or a file-like object.
    validate : bool | None
        Whether to validate the XML document against the OME schema before parsing.
        If None, validation will be skipped if lxml is not available,
        and will be performed otherwise.
    parser_kwargs : ParserKwargs | None
        Passed to the XmlParser constructor. If None, a default parser
        will be used.
    """
    xml = tiff2xml(path)
    return from_xml(xml, validate=validate, parser_kwargs=parser_kwargs)


TIFF_TYPES: dict[bytes, tuple[Struct, Struct, int, Struct]] = {
    b"II*\0": (Struct("<I"), Struct("<H"), 12, Struct("<H")),
    b"MM\0*": (Struct(">I"), Struct(">H"), 12, Struct(">H")),
    b"II+\0": (Struct("<Q"), Struct("<Q"), 20, Struct("<H")),
    b"MM\0+": (Struct(">Q"), Struct(">Q"), 20, Struct(">H")),
}


def _unpack(fh: BinaryIO, strct: Struct) -> int:
    return strct.unpack(fh.read(strct.size))[0]


def tiff2xml(path: Path | str | BinaryIO) -> bytes:
    """Extract the OME-XML from a TIFF file."""
    if hasattr(path, "read"):
        ctx: ContextManager[BinaryIO] = nullcontext(path)  # type: ignore[arg-type]
    else:
        ctx = Path(path).open(mode="rb")

    with ctx as fh:
        head = fh.read(4)
        if head not in TIFF_TYPES:  # pragma: no cover
            raise ValueError(f"{path!r} does not have a recognized TIFF header")

        offset_fmt, tagno_fmt, tagsize, codeformat = TIFF_TYPES[head]
        offset_size = offset_fmt.size
        offset_size_4 = offset_size + 4

        if offset_size == 8:
            fh.seek(4, 1)
        fh.seek(_unpack(fh, offset_fmt))
        for _ in range(_unpack(fh, tagno_fmt)):
            tagstruct = fh.read(tagsize)
            if codeformat.unpack(tagstruct[:2])[0] == 270:
                size = offset_fmt.unpack(tagstruct[4:offset_size_4])[0]
                if size <= offset_size:
                    desc = tagstruct[offset_size_4 : offset_size_4 + size]
                    break
                fh.seek(offset_fmt.unpack(tagstruct[-offset_size:])[0])
                desc = fh.read(size)
                break
        else:  # pragma: no cover
            raise ValueError(f"No OME metadata found in file: {path}")

    if desc[-1] == 0:
        desc = desc[:-1]  # pragma: no cover
    return desc


# ------------------------


def to_dict(source: OME | XMLSource) -> dict[str, Any]:
    """Return a dictionary representation of an OME or XML document.

    Parameters
    ----------
    source : OME | Path | str | bytes | io.BytesIO
        An OME object, or a path to an XML file, or a string or bytes containing XML.

    Returns
    -------
    dict[str, Any]
        A dictionary representation of the OME object or XML document.
    """
    if isinstance(source, BaseModel):
        return source.dict(exclude_defaults=True)

    return from_xml(  # type: ignore[return-value]
        source,
        # the class_factory is what prevents class instantiation,
        # simply returning the params instead
        parser_kwargs={"config": ParserConfig(class_factory=lambda a, b: b)},
    )


def to_xml(
    obj: OMEType,
    *,
    # exclude_defaults takes precedence over exclude_unset
    # if a value equals the default, it will be excluded
    exclude_defaults: bool = False,
    # exclude_unset will exclude any value that is not explicitly set
    # but will INCLUDE values that are set to their default
    exclude_unset: bool = True,
    indent: int = 2,
    include_namespace: bool | None = None,
    include_schema_location: bool = True,
    canonicalize: bool = False,
    validate: bool = False,
) -> str:
    """Generate an XML document from an OME object.

    Parameters
    ----------
    obj : OMEType
        Instance of an ome-types model class.
    exclude_defaults : bool, optional
        Whether to exclude attributes that are set to their default value,
        by default False.
    exclude_unset : bool, optional
        Whether to exclude attributes that are not explicitly set,
        by default True.
    indent : int, optional
        Number of spaces to indent the XML document, by default 2.
    include_namespace : bool | None, optional
        Whether to include the OME namespace in the root element.  If `None`, will
        be set to the value of `canonicalize`, by default None.
    include_schema_location : bool, optional
        Whether to include the schema location in the root element, by default True.
    canonicalize : bool, optional
        Whether to canonicalize the XML output, by default False.
    validate : bool, optional
        Whether to validate the XML document against the OME schema, after rendering.
        (In most cases, this will be redundant and unnecessary.)

    Returns
    -------
    str
        The XML document as a string.
    """
    config = SerializerConfig(
        pretty_print=(indent > 0) and not canonicalize,  # canonicalize does it for us
        pretty_print_indent=" " * indent,
        xml_declaration=False,
        ignore_default_attributes=exclude_defaults,
        ignore_unset_attributes=exclude_unset,
        attribute_sort_key=operator.attrgetter("name") if canonicalize else None,
    )
    if include_schema_location:
        config.schema_location = f"{OME_2016_06_URI} {OME_2016_06_URI}/ome.xsd"

    serializer = XmlSerializer(config=config)
    if include_namespace is None:
        include_namespace = canonicalize

    if exclude_unset:
        # if we're excluding unset attributes, we need to be very careful that the
        # __fields_set__ attribute is accurate, otherwise we'll exclude attributes
        # this is tricky for things like mutable sequences that pydantic doesn't
        # know about. this method recurses the object and updates the __fields_set__
        # attribute if the field is not equal to its default value
        obj._update_set_fields()

    ns_map = {"ome" if include_namespace else None: OME_2016_06_URI}
    xml = serializer.render(obj, ns_map=ns_map)

    if canonicalize:
        xml = _canonicalize(xml, indent=" " * indent)
    if validate:
        validate_xml(xml)
    return xml


def _canonicalize(xml: str, indent: str) -> str:
    from xml.dom import minidom

    xml_out = ET.canonicalize(xml, strip_text=True)
    return minidom.parseString(xml_out).toprettyxml(indent=indent)


# ------------------------


class ValidationError(ValueError):
    ...


def validate_xml(
    xml: XMLSource, schema: Path | str | None = None, *, warn_on_transform: bool = True
) -> ET._ElementTree:
    """Validate XML against an XML Schema.

    By default, will validate against the OME 2016-06 schema.
    """
    with suppress(ImportError):
        return validate_xml_with_lxml(xml, schema, warn_on_transform)

    with suppress(ImportError):  # pragma: no cover
        return validate_xml_with_xmlschema(xml, schema, warn_on_transform)

    raise ImportError(  # pragma: no cover
        "Validation requires either `lxml` or `xmlschema`. "
        "Please pip install one of them."
    ) from None


def validate_xml_with_lxml(
    xml: XMLSource, schema: Path | str | None = None, warn_on_transform: bool = True
) -> ET._ElementTree:
    """Validate XML against an XML Schema using lxml."""
    from lxml import etree

    tree = ensure_2016(xml, warn_on_transform=warn_on_transform, as_tree=True)
    xmlschema = etree.XMLSchema(etree.parse(schema or OME_2016_06_XSD))

    if not xmlschema.validate(tree):
        msg = f"Validation of {str(xml)[:20]!r} failed:"
        for error in xmlschema.error_log:
            msg += f"\n  - line {error.line}: {error.message}"
        raise ValidationError(msg)
    return tree


def validate_xml_with_xmlschema(
    xml: XMLSource, schema: Path | str | None = None, warn_on_transform: bool = True
) -> ET._ElementTree:
    """Validate XML against an XML Schema using xmlschema."""
    from xmlschema.exceptions import XMLSchemaException

    tree = ensure_2016(xml, warn_on_transform=warn_on_transform, as_tree=True)
    xmlschema = _get_XMLSchema(schema or OME_2016_06_XSD)
    try:
        xmlschema.validate(tree)  # type: ignore[arg-type]
    except XMLSchemaException as e:
        raise ValidationError(str(e)) from None
    return tree


@lru_cache(maxsize=None)
def _get_XMLSchema(schema: Path | str) -> xmlschema.XMLSchema:
    import xmlschema

    xml_schema = xmlschema.XMLSchema(schema)
    # FIXME Hack to work around xmlschema poor support for keyrefs to
    # substitution groups
    ls_sgs = xml_schema.maps.substitution_groups[f"{OME_2016_06_NS}LightSourceGroup"]
    ls_id_maps = xml_schema.maps.identities[f"{OME_2016_06_NS}LightSourceIDKey"]
    ls_id_maps.elements = {e: None for e in ls_sgs}
    return xml_schema


# ------------------------

TRANSFORMS_PATH = Path(__file__).parent / "transforms"
TRANSFORMS = {
    f"{OME_ROOT}/{p.name.split('-to')[0]}": p for p in TRANSFORMS_PATH.glob("*.xsl")
}


@overload
def ensure_2016(
    source: XMLSource, *, warn_on_transform: bool = ..., as_tree: Literal[True]
) -> ET._ElementTree:
    ...


@overload
def ensure_2016(
    source: XMLSource, *, warn_on_transform: bool = ..., as_tree: Literal[False] = ...
) -> FileLike:
    ...


def ensure_2016(
    source: XMLSource, *, warn_on_transform: bool = False, as_tree: bool = False
) -> FileLike | ET._ElementTree:
    """Ensure source is OME-2016-06 XML.

    If the source is not OME-2016-06 XML, it will be transformed sequentially using
    XSLT transforms in the `transforms` directory, until it is in the 2016-06 namespace.
    NOTE: this requires lxml to be installed and will raise an ImportError if it is not.

    Parameters
    ----------
    source : Path | str | bytes | io.BytesIO
        Path to an XML file, string or bytes containing XML, or a file-like object.
    warn_on_transform : bool
        Whether to warn if a transformation was applied to bring the document to
        OME-2016-06.
    as_tree : bool
        Whether to return an ElementTree or a FileLike object.

    Returns
    -------
    FileLike | ET._ElementTree
        If `as_tree` is `True`, an ElementTree, otherwise a FileLike object representing
        transformed OME 2016 XML.


    Raises
    ------
    ValueError
        If the source is an unrecognized XML namespace.
    ImportError
        If lxml is not installed and a transformation is required.
    """
    normed_source = _normalize(source)
    ns_in = _get_ns_file(normed_source)

    if ns_in == OME_2016_06_URI:
        if as_tree:
            if hasattr(normed_source, "seek"):
                normed_source.seek(0)
            return ET.parse(normed_source)
        return normed_source

    if ns_in in TRANSFORMS:
        tree = ET.parse(normed_source)
        ns = ns_in
        while ns in TRANSFORMS:
            tree = _apply_xslt(tree, TRANSFORMS[ns])
            ns = _get_ns_elem(tree)
        if warn_on_transform:
            warnings.warn(
                f"Transformed source from {ns_in!r} to {OME_2016_06_URI!r}",
                stacklevel=2,
            )

        return tree if as_tree else io.BytesIO(ET.tostring(tree, encoding="utf-8"))

    raise ValueError(f"Unsupported document namespace {ns!r}")  # pragma: no cover


def _normalize(source: XMLSource) -> FileLike:
    """Normalize the input to a BytesIO object, or a filepath.

    Parameters
    ----------
    source : Path | str | bytes | io.BytesIO
        Path to an XML file, string or bytes containing XML, or a file-like object.

    Returns
    -------
    FileLike
        A BytesIO object, or a filepath.
    """
    if isinstance(source, Path):
        return str(source.resolve())
    elif isinstance(source, str):
        if os.path.isfile(source):
            return source
        return io.BytesIO(source.encode())
    elif isinstance(source, bytes):
        return io.BytesIO(source)
    elif isinstance(source, io.BufferedIOBase):
        return source
    raise TypeError(f"Unsupported source type {type(source)!r}")  # pragma: no cover


def _apply_xslt(root: ElementOrTree, xslt_path: str | Path) -> _XSLTResultTree:
    """Apply an XSLT transform to an element or element tree."""
    try:
        from lxml import etree
    except ImportError:  # pragma: no cover
        ns = _get_ns_elem(root)
        raise ImportError(
            f"This documented is using an outdated schema ({ns!r}). "
            "But lxml is required to update older schemas to OME-2016. "
            "Please run `pip install lxml`"
        ) from None

    # Create a transformer object
    transformer = etree.XSLT(ET.parse(xslt_path))
    return transformer(root)


# ------------------------


def _get_ns_elem(elem: ET._Element | ET._ElementTree) -> str:
    """Get namespace from an element or element tree."""
    root = elem.getroot() if hasattr(elem, "getroot") else elem
    # return root.nsmap[root.prefix]  this only works for lxml
    return root.tag.split("}", 1)[0].lstrip("{")


def _get_ns_file(source: FileLike) -> str:
    """Get namespace from a file or file-like object."""
    _, root = next(ET.iterparse(source, events=("start",)))
    return _get_ns_elem(root)  # type: ignore[arg-type]


def _get_root_ome_type(xml: FileLike | ET._ElementTree) -> type[OMEType]:
    """Resolve a ome_types.model class for the root element of an OME XML document."""
    from ome_types import model

    if hasattr(xml, "getroot"):
        root = xml.getroot()
    else:
        if hasattr(xml, "seek"):
            xml.seek(0)
        _, root = next(ET.iterparse(xml, events=("start",)))
    localname = cast("ET._Element", root).tag.rsplit("}", 1)[-1]

    if hasattr(xml, "seek"):
        xml.seek(0)
    try:
        return getattr(model, localname)
    except AttributeError:
        raise ValueError(f"Unknown root element {localname!r}") from None



from __future__ import annotations

from contextlib import contextmanager
from typing import TYPE_CHECKING, Iterator

from xsdata.codegen.analyzer import ClassAnalyzer
from xsdata.codegen.container import ClassContainer
from xsdata.codegen.handlers import RenameDuplicateAttributes
from xsdata.codegen.mappers.schema import SchemaMapper
from xsdata.codegen.transformer import SchemaTransformer
from xsdata.models.xsd import Attribute

if TYPE_CHECKING:
    from xsdata.codegen.models import Class
    from xsdata.models.xsd import Schema


# we don't need RenameDuplicateAttributes because we inject
# proper enum names in our _generator.py
UNWANTED_HANDLERS = [RenameDuplicateAttributes]


def display_help(self: Attribute) -> str | None:
    """Monkeypatched display_help to also mine simple_type.annotations for docs."""
    help_str = super(Attribute, self).display_help

    if not help_str and self.simple_type is not None:
        help_str = "\n".join(
            doc.tostring() or ""
            for annotation in self.simple_type.annotations
            for doc in annotation.documentations
        ).strip()

    return help_str or None


@contextmanager
def monkeypatched_display_help() -> Iterator[None]:
    prev = Attribute.display_help
    Attribute.display_help = property(display_help)  # type: ignore
    try:
        yield
    finally:
        Attribute.display_help = prev  # type: ignore


class OMESchemaMapper(SchemaMapper):
    # may not need to override this... but here just in case
    pass


class OMETransformer(SchemaTransformer):
    # overriding to use our own schema mapper
    def generate_classes(self, schema: Schema) -> list[Class]:
        """Convert the given schema tree to a list of classes."""
        with monkeypatched_display_help():
            return OMESchemaMapper.map(schema)

    # overriding to remove the certain handlers
    def analyze_classes(self, classes: list[Class]) -> list[Class]:
        """Analyzer the given class list and simplify attributes and extensions."""
        # xsdata makes this particular class hard to extend/modify
        container = ClassContainer(config=self.config)

        for handlers in container.processors.values():
            for h in list(handlers):
                for unwanted in UNWANTED_HANDLERS:
                    if isinstance(h, unwanted):
                        handlers.remove(h)

        container.extend(classes)

        return ClassAnalyzer.process(container)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

try:
    import pint
except ImportError:  # pragma: no cover
    raise ImportError(
        "Pint is required to use quantities in ome-types. "
        "Install with `pip install ome-types[pint]`."
    ) from None


if TYPE_CHECKING:
    from pydantic import BaseModel

# The [`pint.UnitRegistry`][] used by ome-types.
ureg: pint.UnitRegistry = pint.UnitRegistry(auto_reduce_dimensions=True)

ureg.define("reference_frame = [_reference_frame]")
ureg.define("@alias grade = gradian")
ureg.define("@alias astronomical_unit = ua")
ureg.define("line = inch / 12")
ureg.define("millitorr = torr / 1000 = mTorr")
ureg.define("@alias torr = Torr")

_UNIT_FIELD = "{}_unit"


def _quantity_property(field_name: str) -> property:
    """Create property that returns a ``pint.Quantity`` combining value and unit."""

    def quantity(self: Any) -> pint.Quantity | None:
        value = getattr(self, field_name)
        if value is None:  # pragma: no cover
            return None

        unit = getattr(self, _UNIT_FIELD.format(field_name))
        return ureg.Quantity(value, unit.value.replace(" ", "_"))

    return property(quantity)


def add_quantity_properties(cls: type[BaseModel]) -> None:
    """Add quantity properties to each field with a corresponding *_unit field.

    Quantity properties allow you to access the value and unit of a field in a single
    X_quantity property, where X is the name of the field.  It returns a pint object.
    """
    _QUANTITY_FIELD = "{}_quantity"
    for field in cls.__fields__:
        if _UNIT_FIELD.format(field) in cls.__fields__:
            setattr(cls, _QUANTITY_FIELD.format(field), _quantity_property(field))



import weakref
from typing import Any, Dict, Optional, Union

from ome_types._mixins._base_type import OMEType


class ReferenceMixin(OMEType):
    _ref: Optional[weakref.ReferenceType] = None

    @property
    def ref(self) -> Union[OMEType, None]:
        if self._ref is None:
            raise ValueError("references not yet resolved on root OME object")
        return self._ref()

    def __getstate__(self: Any) -> Dict[str, Any]:
        """Support pickle of our weakref references."""
        state = super().__getstate__()
        state["__private_attribute_values__"].pop("_ref", None)
        return state



try:
    from importlib.metadata import PackageNotFoundError, version
except ImportError:  # pragma: no cover
    from importlib_metadata import PackageNotFoundError, version  # type: ignore

try:
    __version__ = version("ome-types")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ome_types.units import ureg  # noqa: TCH004

from ome_types import model
from ome_types._conversion import from_tiff, from_xml, to_dict, to_xml, validate_xml
from ome_types.model import OME

__all__ = [
    "__version__",
    "from_tiff",
    "from_xml",
    "model",
    "OME",
    "to_dict",
    "to_xml",
    "ureg",
    "validate_xml",
]


def __getattr__(name: str) -> Any:
    if name == "ureg":
        from ome_types.units import ureg

        return ureg
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



"""This module is only here for backwards compatibility with v<0.4.0 of ome-types.

Most of the types defined here are also exported from `ome_types.model`.  But for
now, you can also import them from here.
"""
from enum import Enum

from ome_types._autogenerated.ome_2016_06 import (
    Binning,
    Marker,
    NamingConvention,
    PixelType,
    UnitsElectricPotential,
    UnitsFrequency,
    UnitsLength,
    UnitsPower,
    UnitsPressure,
    UnitsTemperature,
    UnitsTime,
)
from ome_types._autogenerated.ome_2016_06.shape import FontFamily
from ome_types.model._color import Color

__all__ = [
    "PixelType",
    "Binning",
    "FontFamily",
    "Hex40",
    "LSID",
    "Marker",
    "Color",
    "NamingConvention",
    "NonNegativeFloat",
    "NonNegativeInt",
    "NonNegativeLong",
    "PercentFraction",
    "PixelType",
    "PositiveFloat",
    "PositiveInt",
    "UnitsAngle",
    "UnitsElectricPotential",
    "UnitsFrequency",
    "UnitsLength",
    "UnitsPower",
    "UnitsPressure",
    "UnitsTemperature",
    "UnitsTime",
    "UniversallyUniqueIdentifier",
    "AnnotationID",
    "ChannelID",
    "DatasetID",
    "DetectorID",
    "DichroicID",
    "ExperimenterGroupID",
    "ExperimenterID",
    "ExperimentID",
    "FilterID",
    "FilterSetID",
    "FolderID",
    "ImageID",
    "InstrumentID",
    "LightSourceID",
    "MicrobeamManipulationID",
    "ModuleID",
    "ObjectiveID",
    "PixelsID",
    "PlateAcquisitionID",
    "PlateID",
    "ProjectID",
    "ReagentID",
    "ROIID",
    "ScreenID",
    "ShapeID",
    "WellID",
    "WellSampleID",
]


Hex40 = bytes
NonNegativeFloat = float
NonNegativeInt = int
NonNegativeLong = int
PercentFraction = float
PositiveFloat = float
PositiveInt = int
UniversallyUniqueIdentifier = str

# IDs

LSID = str
AnnotationID = str
ChannelID = str
DatasetID = str
DetectorID = str
DichroicID = str
ExperimenterGroupID = str
ExperimenterID = str
ExperimentID = str
FilterID = str
FilterSetID = str
FolderID = str
ImageID = str
InstrumentID = str
LightSourceID = str
MicrobeamManipulationID = str
ModuleID = str
ObjectiveID = str
PixelsID = str
ROIID = str
PlateAcquisitionID = str
PlateID = str
ProjectID = str
ReagentID = str
ScreenID = str
ShapeID = str
WellID = str
WellSampleID = str


class UnitsAngle(Enum):
    """The units used to represent an angle."""

    DEGREE = "deg"
    GRADIAN = "gon"
    RADIAN = "rad"



try:
    from lxml import etree as ET
except ImportError:
    import xml.etree.ElementTree as ET  # type: ignore

from dataclasses import MISSING, field
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    cast,
)

from pydantic import BaseModel, validators
from pydantic.fields import Field, ModelField, Undefined
from xsdata.formats.dataclass.compat import Dataclasses, class_types
from xsdata.formats.dataclass.models.elements import XmlType
from xsdata.models.datatype import XmlDate, XmlDateTime, XmlDuration, XmlPeriod, XmlTime

T = TypeVar("T", bound=object)


class AnyElement(BaseModel):
    """Generic model to bind xml document data to wildcard fields.

    :param qname: The element's qualified name
    :param text: The element's text content
    :param tail: The element's tail content
    :param children: The element's list of child elements.
    :param attributes: The element's key-value attribute mappings.
    """

    qname: Optional[str] = Field(default=None)
    text: Optional[str] = Field(default=None)
    tail: Optional[str] = Field(default=None)
    children: List["AnyElement"] = Field(
        default_factory=list, metadata={"type": XmlType.WILDCARD}
    )
    attributes: Dict[str, str] = Field(
        default_factory=dict, metadata={"type": XmlType.ATTRIBUTES}
    )

    class Config:
        arbitrary_types_allowed = True

    def to_etree_element(self) -> "ET._Element":
        elem = ET.Element(self.qname or "", self.attributes)
        elem.text = self.text
        elem.tail = self.tail
        for child in self.children:
            elem.append(child.to_etree_element())
        return elem


class DerivedElement(BaseModel, Generic[T]):
    """Generic model wrapper for type substituted elements.

    Example: eg. <b xsi:type="a">...</b>

    :param qname: The element's qualified name
    :param value: The wrapped value
    :param type: The real xsi:type
    """

    qname: str
    value: T
    type: Optional[str] = None

    class Config:
        arbitrary_types_allowed = True


class PydanticBaseModel(Dataclasses):
    @property
    def any_element(self) -> Type:
        return AnyElement

    @property
    def derived_element(self) -> Type:
        return DerivedElement

    def is_model(self, obj: Any) -> bool:
        clazz = obj if isinstance(obj, type) else type(obj)
        if issubclass(clazz, BaseModel):
            clazz.update_forward_refs()
            return True

        return False

    def get_fields(self, obj: Any) -> Tuple[Any, ...]:
        _fields = cast("BaseModel", obj).__fields__.values()
        return tuple(_pydantic_field_to_dataclass_field(field) for field in _fields)


def _pydantic_field_to_dataclass_field(pydantic_field: ModelField) -> Any:
    if pydantic_field.default_factory is not None:
        default_factory: Any = pydantic_field.default_factory
        default = MISSING
    else:
        default_factory = MISSING
        default = (
            MISSING
            if pydantic_field.default in (Undefined, Ellipsis)
            else pydantic_field.default
        )

    dataclass_field = field(  # type: ignore
        default=default,
        default_factory=default_factory,
        # init=True,
        # hash=None,
        # compare=True,
        metadata=pydantic_field.field_info.extra.get("metadata", {}),
        # kw_only=MISSING,
    )
    dataclass_field.name = pydantic_field.name
    dataclass_field.type = pydantic_field.type_
    return dataclass_field


class_types.register("pydantic-basemodel", PydanticBaseModel())


def make_validators(tp: Type, factory: Callable) -> List[Callable]:
    def validator(value: Any) -> Any:
        if isinstance(value, tp):
            return value

        if isinstance(value, str):
            return factory(value)

        raise ValueError

    return [validator]


if hasattr(validators, "_VALIDATORS"):
    validators._VALIDATORS.extend(
        [
            (XmlDate, make_validators(XmlDate, XmlDate.from_string)),
            (XmlDateTime, make_validators(XmlDateTime, XmlDateTime.from_string)),
            (XmlTime, make_validators(XmlTime, XmlTime.from_string)),
            (XmlDuration, make_validators(XmlDuration, XmlDuration)),
            (XmlPeriod, make_validators(XmlPeriod, XmlPeriod)),
            (ET.QName, make_validators(ET.QName, ET.QName)),
        ]
    )
else:
    import warnings

    warnings.warn(
        "Could not find pydantic.validators._VALIDATORS."
        "xsdata-pydantic-basemodel may be incompatible with your pydantic version.",
        stacklevel=2,
    )



from typing import Iterable, List, MutableSequence, TypeVar, Union, overload

T = TypeVar("T")


class UserSequence(MutableSequence[T]):
    """Generric Mutable sequence, that expects the real list at __root__."""

    __root__: List[object]

    def __repr__(self) -> str:
        return repr(self.__root__)

    def __delitem__(self, _idx: Union[int, slice]) -> None:
        del self.__root__[_idx]

    @overload
    def __getitem__(self, _idx: int) -> T:
        ...

    @overload
    def __getitem__(self, _idx: slice) -> List[T]:
        ...

    def __getitem__(self, _idx: Union[int, slice]) -> Union[T, List[T]]:
        return self.__root__[_idx]  # type: ignore[return-value]

    def __len__(self) -> int:
        return len(self.__root__)

    @overload
    def __setitem__(self, _idx: int, _val: T) -> None:
        ...

    @overload
    def __setitem__(self, _idx: slice, _val: Iterable[T]) -> None:
        ...

    def __setitem__(self, _idx: Union[int, slice], _val: Union[T, Iterable[T]]) -> None:
        self.__root__[_idx] = _val  # type: ignore[index]

    def insert(self, index: int, value: T) -> None:
        self.__root__.insert(index, value)

    # for some reason, without overloading this... append() adds things to the
    # beginning of the list instead of the end
    def append(self, value: T) -> None:
        self.__root__.append(value)



from __future__ import annotations

import os
import re
from collections import defaultdict
from contextlib import contextmanager
from functools import lru_cache
from pathlib import Path
from typing import Any, Iterator, NamedTuple, cast
from xml.etree import ElementTree as ET

SRC_PATH = Path(__file__).parent.parent
SCHEMA_FILE = (SRC_PATH / "ome_types" / "ome-2016-06.xsd").absolute()


@contextmanager
def cd(new_path: str | Path) -> Iterator[None]:
    """Temporarily change the current working directory.

    Used as a workaround for xsdata not supporting output path.
    """
    prev = Path.cwd()
    os.chdir(Path(new_path).expanduser().absolute())
    try:
        yield
    finally:
        os.chdir(prev)


class EnumInfo(NamedTuple):
    name: str
    plural: str
    unitsystem: str
    enum: str


class AppInfo(NamedTuple):
    plurals: dict[str, str]
    enums: dict[str, dict[str, EnumInfo]]
    abstract: list[str]


@lru_cache(maxsize=None)
def get_appinfo(schema: Path | str = SCHEMA_FILE) -> AppInfo:
    """Gather all the <xsd:appinfo> stuff from the schema.

    xsdata doesn't try to do anything with it. But we want to use it to
    provide better enum and plural names
    """
    tree = ET.parse(schema)
    plurals: dict[str, str] = {}
    enums: defaultdict[str, dict[str, EnumInfo]] = defaultdict(dict)
    in_name = ""
    in_value = ""
    abstract = []
    for node in tree.iter():
        if node.tag == "plural":  # <plural>
            plurals[in_name] = cast(str, node.text)
        elif node.tag == "enum":  # <enum>
            enums[in_name][in_value] = EnumInfo(
                node.get("name", ""),
                node.get("plural", ""),
                node.get("unitsystem", ""),
                node.get("enum", ""),
            )
        elif node.tag == "abstract":  # <abstract>
            abstract.append(in_name)
        else:
            in_name = node.get("name", in_name)
            in_value = node.get("value", in_value)

    return AppInfo(plurals, dict(enums), abstract)


CAMEL_SNAKE_OVERRIDES = {"ROIs": "rois"}
camel_snake_registry: dict[str, str] = {}


def camel_to_snake(name: str, **kwargs: Any) -> str:
    """Variant of camel_to_snake that preserves adjacent uppercase letters.

    https://stackoverflow.com/a/1176023
    """
    name = name.lstrip("@")  # remove leading @ from "@any_element"
    result = CAMEL_SNAKE_OVERRIDES.get(name)
    if not result:
        result = re.sub("([A-Z]+)([A-Z][a-z]+)", r"\1_\2", name)
        result = re.sub("([a-z0-9])([A-Z])", r"\1_\2", result)
        result = result.lower().replace(" ", "_")
    camel_snake_registry[name] = result
    return result



"""These validators are added to the generated classes in ome_types._autogenerated.

that logic is in the `methods` method in ome_autogen/_generator.py
"""
import warnings
from typing import TYPE_CHECKING, Any, Dict

if TYPE_CHECKING:
    from ome_types.model import (  # type: ignore
        BinData,
        Pixels,
        PixelType,
        XMLAnnotation,
    )
    from xsdata_pydantic_basemodel.compat import AnyElement


# @root_validator(pre=True)
def bin_data_root_validator(cls: "BinData", values: dict) -> Dict[str, Any]:
    # This catches the case of <BinData Length="0"/>, where the parser may have
    # omitted value from the dict, and sets value to b""
    # seems like it could be done in a default_factory, but that would
    # require more modification of xsdata I think
    if "value" not in values:
        if values.get("length") != 0:  # pragma: no cover
            warnings.warn(
                "BinData length is non-zero but value is missing", stacklevel=2
            )
        values["value"] = b""
    return values


# @root_validator(pre=True)
def pixels_root_validator(cls: "Pixels", values: dict) -> dict:
    if "metadata_only" in values:
        if isinstance(values["metadata_only"], bool):
            if not values["metadata_only"]:
                values.pop("metadata_only")
            else:
                # type ignore in case the autogeneration hasn't been built
                from ome_types.model import MetadataOnly  # type: ignore

                values["metadata_only"] = MetadataOnly()

    return values


# @validator("any_elements", each_item=True)
def any_elements_validator(cls: "XMLAnnotation.Value", v: Any) -> "AnyElement":
    # This validator is used because XMLAnnotation.Value.any_elements is
    # annotated as List[object]. So pydantic won't coerce dicts to AnyElement
    # automatically (which is important when constructing OME objects from dicts)
    if isinstance(v, dict):
        # this needs to be delayed until runtime because of circular imports
        from xsdata_pydantic_basemodel.compat import AnyElement

        return AnyElement(**v)
    return v


# @validator('value', pre=True)
def xml_value_validator(cls: "XMLAnnotation", v: Any) -> "XMLAnnotation.Value":
    if isinstance(v, str):
        # FIXME: this is a hack to support passing a string to XMLAnnotation.value
        # there must be a more direct way to do this...
        # (the type ignores here are because the model might not be built yet)
        from ome_types._conversion import OME_2016_06_URI
        from ome_types.model import XMLAnnotation  # type: ignore
        from xsdata_pydantic_basemodel.bindings import XmlParser

        template = '<XMLAnnotation xmlns="{}"><Value>{}</Value></XMLAnnotation>'
        xml = template.format(OME_2016_06_URI, v)
        return XmlParser().from_string(xml, XMLAnnotation).value  # type: ignore
    return v


def pixel_type_to_numpy_dtype(self: "PixelType") -> str:
    """Get a numpy dtype string for this pixel type."""
    m = {
        "float": "float32",
        "double": "float64",
        "complex": "complex64",
        "double-complex": "complex128",
        "bit": "bool",  # ?
    }
    return m.get(self.value, self.value)



from contextlib import suppress
from typing import Tuple, Union

from pydantic import color

__all__ = ["Color"]

RGBA = Tuple[int, int, int, float]
ColorType = Union[Tuple[int, int, int], RGBA, str, int]


class Color(color.Color):
    """A Pydantic Color subclass that converts to and from OME int32 types."""

    def __init__(self, val: ColorType = -1) -> None:
        with suppress(ValueError, TypeError):
            val = self._int2tuple(int(val))  # type: ignore
        super().__init__(val)  # type: ignore [arg-type]

    @classmethod
    def _int2tuple(cls, val: int) -> RGBA:
        return (val >> 24 & 255, val >> 16 & 255, val >> 8 & 255, (val & 255) / 255)

    def as_int32(self) -> int:
        """Convert to an int32, with alpha in the least significant byte."""
        r, g, b, *a = self.as_rgb_tuple()
        v = r << 24 | g << 16 | b << 8 | int((a[0] if a else 1) * 255)
        if v < 2**32 // 2:
            return v
        return v - 2**32

    def __eq__(self, o: object) -> bool:
        if isinstance(o, Color):
            return self.as_int32() == o.as_int32()
        return NotImplemented  # pragma: no cover

    def __int__(self) -> int:
        return self.as_int32()



from typing import TYPE_CHECKING, List, Union, cast

if TYPE_CHECKING:
    from ome_types._autogenerated.ome_2016_06 import (
        Arc,
        Filament,
        GenericExcitationSource,
        Instrument,
        Laser,
        LightEmittingDiode,
    )

    LightSource = Union[
        GenericExcitationSource, LightEmittingDiode, Filament, Arc, Laser
    ]


class InstrumentMixin:
    @property
    def light_source_group(self) -> List["LightSource"]:
        # here for backwards compatibility
        slf = cast("Instrument", self)
        return [
            *slf.arcs,
            *slf.filaments,
            *slf.generic_excitation_sources,
            *slf.lasers,
            *slf.light_emitting_diodes,
        ]



from __future__ import annotations

import os
import subprocess
import sys
from pathlib import Path
from shutil import rmtree

from ome_autogen import _util
from ome_autogen._config import get_config
from ome_autogen._transformer import OMETransformer

BLACK_LINE_LENGTH = 88
BLACK_TARGET_VERSION = "py37"
BLACK_SKIP_TRAILING_COMMA = False  # use trailing commas as a reason to split lines?
OUTPUT_PACKAGE = "ome_types._autogenerated.ome_2016_06"
DO_MYPY = os.environ.get("OME_AUTOGEN_MYPY", "0") == "1" or "--mypy" in sys.argv
SRC_PATH = Path(__file__).parent.parent
SCHEMA_FILE = (SRC_PATH / "ome_types" / "ome-2016-06.xsd").absolute()
RUFF_IGNORE: list[str] = [
    "D101",  # Missing docstring in public class
    "D106",  # Missing docstring in public nested class
    "D205",  # 1 blank line required between summary line and description
    "D404",  # First word of the docstring should not be This
    "E501",  # Line too long
    "S105",  # Possible hardcoded password
]


def build_model(
    output_dir: Path | str = SRC_PATH,
    schema_file: Path | str = SCHEMA_FILE,
    target_package: str = OUTPUT_PACKAGE,
    ruff_ignore: list[str] = RUFF_IGNORE,
    do_formatting: bool = True,
    do_mypy: bool = DO_MYPY,
) -> None:
    """Convert the OME schema to a python model."""
    config = get_config(target_package)
    transformer = OMETransformer(print=False, config=config)

    _print_gray(f"Processing {getattr(schema_file ,'name', schema_file)}...")
    transformer.process_sources([Path(schema_file).resolve().as_uri()])

    package_dir = str(Path(output_dir) / OUTPUT_PACKAGE.replace(".", "/"))
    rmtree(package_dir, ignore_errors=True)
    with _util.cd(output_dir):  # xsdata doesn't support output path
        _print_gray("Writing Files...")
        transformer.process_classes()

    if do_formatting:
        _fix_formatting(package_dir, ruff_ignore)

    if do_mypy:
        _check_mypy(package_dir)

    _print_green(f"OME python model created at {OUTPUT_PACKAGE}")


def _fix_formatting(package_dir: str, ruff_ignore: list[str] = RUFF_IGNORE) -> None:
    _print_gray("Running black and ruff ...")

    ruff = ["ruff", "-q", "--fix", package_dir]
    ruff.extend(f"--ignore={ignore}" for ignore in ruff_ignore)
    subprocess.check_call(ruff)  # noqa S

    black = [
        "black",
        "-q",
        f"--line-length={BLACK_LINE_LENGTH}",
        f"--target-version={BLACK_TARGET_VERSION}",
    ]
    if BLACK_SKIP_TRAILING_COMMA:  # pragma: no cover
        black.append("--skip-magic-trailing-comma")
    black.extend([str(x) for x in Path(package_dir).rglob("*.py")])
    subprocess.check_call(black)  # noqa S


def _check_mypy(package_dir: str) -> None:
    _print_gray("Running mypy ...")

    mypy = ["mypy", package_dir, "--strict"]
    try:
        subprocess.check_output(mypy, stderr=subprocess.STDOUT)  # noqa S
    except subprocess.CalledProcessError as e:  # pragma: no cover
        raise RuntimeError(f"mypy errors:\n\n{e.output.decode()}") from e


def _print_gray(text: str) -> None:
    if os.name != "nt":
        # UnicodeEncodeError: 'charmap' codec can't encode character '\u2713'
        text = f"\033[90m\033[1m{text}\033[0m"
    print(text)


def _print_green(text: str) -> None:
    if os.name != "nt":
        # UnicodeEncodeError: 'charmap' codec can't encode character '\u2713'
        text = f"\033[92m\033[1m{text}\033[0m"
    print(text)



import warnings
from datetime import datetime
from enum import Enum
from textwrap import indent
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    MutableSequence,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    cast,
)

from pydantic import BaseModel, validator

from ome_types._mixins._ids import validate_id

try:
    from ome_types.units import add_quantity_properties
except ImportError:
    add_quantity_properties = lambda cls: None  # noqa: E731

if TYPE_CHECKING:
    from ome_types._conversion import XMLSource


T = TypeVar("T", bound="OMEType")
# Default value to support automatic numbering for id field values.
AUTO_SEQUENCE = "__auto_sequence__"


DEPRECATED_NAMES = {
    "annotation_ref": "annotation_refs",
    "bin_data": "bin_data_blocks",
    "dataset_ref": "dataset_refs",
    "emission_filter_ref": "emission_filters",
    "excitation_filter_ref": "excitation_filters",
    "experimenter_ref": "experimenter_refs",
    "folder_ref": "folder_refs",
    "image_ref": "image_refs",
    "leader": "leaders",
    "light_source_settings": "light_source_settings_combinations",
    "m": "ms",
    "microbeam_manipulation_ref": "microbeam_manipulation_refs",
    "plate_ref": "plate_refs",
    "roi_ref": "roi_refs",
    "well_sample_ref": "well_sample_refs",
}


def _move_deprecated_fields(data: Dict[str, Any], field_names: Set[str]) -> None:
    for key in list(data):
        if (
            key not in field_names
            and key in DEPRECATED_NAMES
            and DEPRECATED_NAMES[key] in field_names
        ):
            warnings.warn(
                f"Field {key!r} is deprecated. Use {DEPRECATED_NAMES[key]!r} instead.",
                stacklevel=2,
            )
            data[DEPRECATED_NAMES[key]] = data.pop(key)


class OMEType(BaseModel):
    """The base class that all OME Types inherit from.

    This provides some global conveniences around auto-setting ids. (i.e., making them
    optional in the class constructor, but never `None` after initialization.).
    It provides a nice `__repr__` that hides things that haven't been changed from
    defaults.  It adds ``*_quantity`` property for fields that have both a value and a
    unit, where `*_quantity` is a [`pint.Quantity`][].  It also provides pickling
    support.
    """

    # pydantic BaseModel configuration.
    # see: https://pydantic-docs.helpmanual.io/usage/model_config/
    class Config:
        arbitrary_types_allowed = False
        validate_assignment = True
        underscore_attrs_are_private = True
        use_enum_values = False
        validate_all = True

    # allow use with weakref
    __slots__: ClassVar[Set[str]] = {"__weakref__"}  # type: ignore

    _v = validator("id", pre=True, always=True, check_fields=False)(validate_id)

    def __init__(self, **data: Any) -> None:
        warn_extra = data.pop("warn_extra", True)
        field_names = set(self.__fields__.keys())
        _move_deprecated_fields(data, field_names)
        super().__init__(**data)
        kwargs = set(data.keys())
        extra = kwargs - field_names
        if extra and warn_extra:
            warnings.warn(
                f"Unrecognized fields for type {type(self)}: {kwargs - field_names}",
                stacklevel=2,
            )

    def __init_subclass__(cls) -> None:
        """Add `*_quantity` property for fields that have both a value and a unit.

        where `*_quantity` is a pint `Quantity`.
        """
        add_quantity_properties(cls)

    def __repr_args__(self) -> Sequence[Tuple[Optional[str], Any]]:
        """Repr with only set values, and truncated sequences."""
        args = []
        for k, v in self._iter(exclude_defaults=True):
            if isinstance(v, Sequence) and not isinstance(v, str):
                if v == []:  # skip empty lists
                    continue
                # if this is a sequence with a long repr, just show the length
                # and type
                if len(repr(v).split(",")) > 5:
                    type_name = self.__fields__[k].type_.__name__
                    v = _RawRepr(f"[<{len(v)} {type_name}>]")
            elif isinstance(v, Enum):
                v = v.value
            elif isinstance(v, datetime):
                v = v.isoformat()
            args.append((k, v))
        return sorted(args, key=lambda f: f[0] not in ("name", "id"))

    def __repr__(self) -> str:
        lines = [f"{key}={val!r}," for key, val in self.__repr_args__()]
        if len(lines) == 1:
            body = lines[-1].rstrip(",")
        elif lines:
            body = "\n" + indent("\n".join(lines), "   ") + "\n"
        else:
            body = ""
        return f"{self.__class__.__qualname__}({body})"

    def __getattr__(self, key: str) -> Any:
        """Getattr that redirects deprecated names."""
        cls_name = self.__class__.__name__
        if key in DEPRECATED_NAMES and hasattr(self, DEPRECATED_NAMES[key]):
            new_key = DEPRECATED_NAMES[key]
            warnings.warn(
                f"Attribute '{cls_name}.{key}' is deprecated, use {new_key!r} instead",
                DeprecationWarning,
                stacklevel=2,
            )
            return getattr(self, new_key)
        raise AttributeError(f"{cls_name} object has no attribute {key!r}")

    def to_xml(self, **kwargs: Any) -> str:
        """Serialize this object to XML.

        See docstring of [`ome_types.to_xml`][] for kwargs.
        """
        from ome_types._conversion import to_xml

        return to_xml(self, **kwargs)

    @classmethod
    def from_xml(cls: Type[T], xml: "XMLSource", **kwargs: Any) -> T:
        """Read an ome-types class from XML.

        See docstring of [`ome_types.from_xml`][] for kwargs.

        Note: this will return an instance of whatever the top node is in the XML.
        so technically, the return type here could be incorrect.  But when used
        properly (`Image.from_xml()`, `Pixels.from_xml()`, etc.) on an unusual xml
        document it can provide additional typing information.
        """
        from ome_types._conversion import from_xml

        return cast(T, from_xml(xml, **kwargs))

    def _update_set_fields(self) -> None:
        """Update set fields with populated mutable sequences.

        Because pydantic isn't aware of mutations to sequences, it can't tell when
        a field has been "set" by mutating a sequence.  This method updates the
        self.__fields_set__ attribute to reflect that.  We assume that if an attribute
        is not None, and is not equal to the default value, then it has been set.
        """
        for field_name, field in self.__fields__.items():
            current = getattr(self, field_name)
            if not current:
                continue
            if current != field.get_default():
                self.__fields_set__.add(field_name)
            if isinstance(current, OMEType):
                current._update_set_fields()
            if isinstance(current, MutableSequence):
                for item in current:
                    if isinstance(item, OMEType):
                        item._update_set_fields()


class _RawRepr:
    """Helper class to allow repr to show raw values for fields that are sequences."""

    def __init__(self, raw: str) -> None:
        self.raw = raw

    def __repr__(self) -> str:
        return self.raw



from contextlib import suppress
from typing import Dict, Iterator, List, Type, Union

from pydantic import Field, ValidationError, validator

from ome_types._autogenerated.ome_2016_06.ellipse import Ellipse
from ome_types._autogenerated.ome_2016_06.label import Label
from ome_types._autogenerated.ome_2016_06.line import Line
from ome_types._autogenerated.ome_2016_06.mask import Mask
from ome_types._autogenerated.ome_2016_06.point import Point
from ome_types._autogenerated.ome_2016_06.polygon import Polygon
from ome_types._autogenerated.ome_2016_06.polyline import Polyline
from ome_types._autogenerated.ome_2016_06.rectangle import Rectangle
from ome_types._mixins._base_type import OMEType
from ome_types.model._user_sequence import UserSequence

ShapeType = Union[Rectangle, Mask, Point, Ellipse, Line, Polyline, Polygon, Label]
_KINDS: Dict[str, Type[ShapeType]] = {
    "rectangle": Rectangle,
    "mask": Mask,
    "point": Point,
    "ellipse": Ellipse,
    "line": Line,
    "polyline": Polyline,
    "polygon": Polygon,
    "label": Label,
}

_ShapeCls = tuple(_KINDS.values())


class ShapeUnion(OMEType, UserSequence[ShapeType]):  # type: ignore[misc]
    """A mutable sequence of [`ome_types.model.Shape`][].

    Members of this sequence must be one of the following types:

    - [`ome_types.model.Rectangle`][]
    - [`ome_types.model.Mask`][]
    - [`ome_types.model.Point`][]
    - [`ome_types.model.Ellipse`][]
    - [`ome_types.model.Line`][]
    - [`ome_types.model.Polyline`][]
    - [`ome_types.model.Polygon`][]
    - [`ome_types.model.Label`][]
    """

    # NOTE: in reality, this is List[ShapeGroupType]... but
    # for some reason that messes up xsdata data binding
    __root__: List[object] = Field(
        default_factory=list,
        metadata={
            "type": "Elements",
            "choices": tuple(
                {"name": kind.title(), "type": cls} for kind, cls in _KINDS.items()
            ),
        },
    )

    @validator("__root__", each_item=True)
    def _validate_root(cls, v: ShapeType) -> ShapeType:
        if isinstance(v, _ShapeCls):
            return v
        if isinstance(v, dict):
            # NOTE: this is here to preserve the v1 behavior of passing a dict like
            # {"kind": "label", "x": 0, "y": 0}
            # to create a label rather than a point
            if "kind" in v:
                kind = v.pop("kind").lower()
                return _KINDS[kind](**v)

            for cls_ in _ShapeCls:
                with suppress(ValidationError):
                    return cls_(warn_extra=False, **v)
        raise ValueError(f"Invalid shape: {v}")  # pragma: no cover

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__root__!r})"

    # overriding BaseModel.__iter__ to behave more like a real Sequence
    def __iter__(self) -> Iterator[ShapeType]:  # type: ignore[override]
        yield from self.__root__  # type: ignore[misc]  # see NOTE above

    def __eq__(self, _value: object) -> bool:
        return _value == self.__root__
